Attribute VB_Name = "modMain"
' /*******************************************************************************
'   modMain.bas within vbSpec.vbp
'
'   Public variable declarations, startup and initialization code,
'   and routines for loading the ROM and snapshots
'
'   Author: Chris Cowley <ccowley@grok.co.uk>
'
'   Copyright (C)1999-2000 Grok Developments Ltd.
'   http://www.grok.co.uk/
'
'   This program is free software; you can redistribute it and/or
'   modify it under the terms of the GNU General Public License
'   as published by the Free Software Foundation; either version 2
'   of the License, or (at your option) any later version.
'   This program is distributed in the hope that it will be useful,
'   but WITHOUT ANY WARRANTY; without even the implied warranty of
'   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
'   GNU General Public License for more details.
'
'   You should have received a copy of the GNU General Public License
'   along with this program; if not, write to the Free Software
'   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
'
' *******************************************************************************/

Option Explicit

Public glTstatesPerInterrupt As Long
Public Parity(256) As Boolean
Public glInterruptTimer As Long

' // Memory haindling variables for 128K spectrum emulation
Public glMemPagingType As Long ' // 0 = No mem paging (48K speccy)
                               ' // 1 = 128K/+2 paging available
                               ' // 2 = 128K/+2 and +2A/+3 special paging available
Public gRAMPage(11, 16383) As Byte ' // (pages 0 to 7 are RAM, and pages 8 to 11 are ROM)
Public glPageAt(4) As Long
Public glLastOut7FFD As Long
Public glLastOut1FFD As Long
Public glUseScreen As Long
Public glEmulatedModel As Long

#If USE_WINAPI Then
    Public gbSoundEnabled As Boolean
    Public glSoundRegister As Long  ' // Contains currently indexed AY-3-8912 sound register
    Public glBufNum As Long         ' // ID of the last Wave buffer used
#End If

' // Main Z80 registers //
Public regA As Long
Public regHL As Long
Public regB As Long
Public regC As Long
Public regDE As Long

' // Z80 Flags
Public fS As Boolean
Public fZ As Boolean
Public f5 As Boolean
Public fH As Boolean
Public f3 As Boolean
Public fPV As Boolean
Public fN As Boolean
Public fC As Boolean

' // Flag bit positions
Public Const F_C = 1
Public Const F_N = 2
Public Const F_PV = 4
Public Const F_3 = 8
Public Const F_H = 16
Public Const F_5 = 32
Public Const F_Z = 64
Public Const F_S = 128

' // Alternate registers //
Public regAF_ As Long
Public regHL_ As Long
Public regBC_ As Long
Public regDE_ As Long

' // Index registers  - ID used as temp for ix/iy
Public regIX As Long
Public regIY As Long
Public regID As Long

' // Stack pointer and program counter
Public regSP As Long
Public regPC As Long

' // Interrupt registers and flip-flops, and refresh registers
Public intI As Long
Public intR As Long
Public intRTemp As Long
Public intIFF1 As Boolean
Public intIFF2 As Boolean
Public intIM As Long

' // Addressable Memory (16KB ROM at 0x0000, 48KB RAM at 0x4000)
' // We also place a shadow copy of the ROM above the addressable memory space
' // at 0x10000 so we don't have to check for regPC > 65535 at each instruction
Public mem(81919) As Byte

' // Screen things
Public Const nPixelsWide = 256
Public Const nPixelsHigh = 192
Public Const nCharsWide = 32
Public Const nCharsHigh = 24
Public Const FirstAttr = nPixelsHigh * nCharsWide
Public Const lastAddr = FirstAttr + (nCharsHigh * nCharsWide)

Public ScrnLines(nPixelsHigh - 1, 32) As Boolean
Public ScrnNeedRepaint As Boolean
Public bFlashInverse As Boolean
Public glScreenMem(191, 31) As Long

' // Array of colour values (speeds up screen painting by avoiding
' // multiple calls to RGB() )
Public glBrightColor(0 To 7) As Long
Public glNormalColor(0 To 7) As Long

' // Global picDisplay variable to speed things up
Public gpicDisplay As PictureBox

' // Interrupts/Screen refreshing
Public interruptCounter As Long
Public glInterruptDelay As Long
Public glDelayOverage As Long

' // Keypresses
Public keyB_SPC As Long
Public keyH_ENT As Long
Public keyY_P As Long
Public key6_0 As Long
Public key1_5 As Long
Public keyQ_T As Long
Public keyA_G As Long
Public keyCAPS_V As Long

' // Sadly, I needed to use these high res timer functions to precisely control the
' // speed of emulation. I had hoped to do this without resorting to any API calls :(
Public Declare Function timeGetTime Lib "winmm.dll" () As Long
Public Declare Function timeBeginPeriod Lib "winmm.dll" (ByVal uPeriod As Long) As Long
Public Declare Function timeEndPeriod Lib "winmm.dll" (ByVal uPeriod As Long) As Long
Public Declare Sub Sleep Lib "kernel32" (ByVal dwMilliseconds As Long)

' // If built with the USE_WINAPI compiler directive defined, vbSpec uses Windows API
' // functions to paint the display. This is faster than raw VB code, and provides the
' // option for selecting scaled display sizes (double and triple size, and so on).
#If USE_WINAPI Then
    Public Const SRCCOPY = &HCC0020 ' (DWORD) dest = source
    Public Declare Function StretchDIBits Lib "gdi32" (ByVal hdc As Long, ByVal X As Long, ByVal Y As Long, ByVal dx As Long, ByVal dy As Long, ByVal SrcX As Long, ByVal SrcY As Long, ByVal wSrcWidth As Long, ByVal wSrcHeight As Long, lpBits As Any, lpBitsInfo As BITMAPINFO, ByVal wUsage As Long, ByVal dwRop As Long) As Long
    Public Declare Function GetSystemMetrics Lib "user32" (ByVal nIndex As Long) As Long
    Public Const SM_CYCAPTION = 4
    Public Const SM_CYMENU = 15
    Public Const SM_CXFRAME = 32
    Public Const SM_CYFRAME = 33
    
    Public Type BITMAPINFOHEADER '40 bytes
            biSize As Long
            biWidth As Long
            biHeight As Long
            biPlanes As Integer
            biBitCount As Integer
            biCompression As Long
            biSizeImage As Long
            biXPelsPerMeter As Long
            biYPelsPerMeter As Long
            biClrUsed As Long
            biClrImportant As Long
    End Type
    Public Type RGBQUAD
            rgbBlue As Byte
            rgbGreen As Byte
            rgbRed As Byte
            rgbReserved As Byte
    End Type
    Public Type BITMAPINFO
            bmiHeader As BITMAPINFOHEADER
            bmiColors(16) As RGBQUAD
    End Type
    Public Const BI_RGB = 0&
    Public Const DIB_RGB_COLORS = 0 '  color table in RGBs
    
    Public bmiBuffer As BITMAPINFO
    Public cBufferBits(49152) As Byte
    
    Public glDisplayHeight As Long
    Public glDisplayWidth As Long
    Public glDisplayVSource As Long ' // Set to glDisplayHeight - 1 to improve display speed
    Public glDisplayVSize As Long   ' // Set to -glDisplayHeight to improve display speed
    Public glDisplayXMultiplier As Long
    Public glDisplayYMultiplier As Long
    
    ' // Hook the window procedure to process messages for the waveOut() sound emulation
    Public Declare Function SetWindowLong Lib "user32" Alias "SetWindowLongA" (ByVal hwnd As Long, ByVal nIndex As Long, ByVal dwNewLong As Long) As Long
    Public Const GWL_WNDPROC = (-4)
    Public glPrevWndProc As Long ' // Pointer to the 'real' Window Procedure
    Public glhWnd As Long
    
    Public glBeeperVal As Long
#End If

' // ShellExecute is used for the clickable web URL in the "About..." dialog, not
' //the actual emulator itself
Public Declare Function ShellExecute Lib "shell32.dll" Alias "ShellExecuteA" (ByVal hwnd As Long, ByVal lpOperation As String, ByVal lpFile As String, ByVal lpParameters As String, ByVal lpDirectory As String, ByVal nShowCmd As Long) As Long



Public Sub CloseWaveOut()
    #If USE_WINAPI Then
        Dim lRet As Long
        
        lRet = waveOutReset(glphWaveOut)
        For lRet = 1 To NUM_WAV_BUFFERS
            waveOutUnprepareHeader glphWaveOut, gtWavHdr(lRet), Len(gtWavHdr(lRet))
        Next lRet
        
        For lRet = 1 To NUM_WAV_BUFFERS
            GlobalUnlock ghMem(lRet)
            GlobalFree ghMem(lRet)
        Next lRet
        
        waveOutClose glphWaveOut
        
        gbSoundEnabled = False
    #End If
End Sub

Private Function CompressMemoryBlock(lBlock As Long, sData As String) As Long
    Dim lCounter As Long, lCounter2 As Long, lByteCount As Long
    
    Do
        If gRAMPage(lBlock, lCounter) = gRAMPage(lBlock, lCounter + 1) Then
            ' // String of two or more bytes starting from lCounter
            ' // Might be a candidate for compression
            lCounter2 = lCounter
            Do While (lCounter2 - lCounter) < 260
                If gRAMPage(lBlock, lCounter2 + 1) = gRAMPage(lBlock, lCounter2) Then
                    lCounter2 = lCounter2 + 1
                Else
                    Exit Do
                End If
            Loop
            lByteCount = lCounter2 - lCounter + 1
            lCounter = lCounter2 + 1
            If lByteCount > 255 Then
                lCounter = lCounter - (lByteCount - 255)
                lByteCount = 255
            End If
            
            If lByteCount > 4 Then
                ' // Strings of 5 or more bytes are encoded, and the byte following the encoded
                ' // sequence is not counted in any subsequent blocks
                sData = sData & Chr$(&HED) & Chr$(&HED) & Chr$(lByteCount) & Chr$(gRAMPage(lBlock, lCounter - 1))
                If lCounter < 16384 Then
                    sData = sData & Chr$(gRAMPage(lBlock, lCounter))
                    lCounter = lCounter + 1
                End If
            Else
                If gRAMPage(lBlock, lCounter - 1) = &HED& Then
                    ' // Strings of 2 or more ED bytes are always encoded
                    sData = sData & Chr$(&HED) & Chr$(&HED) & Chr$(lByteCount) & Chr$(gRAMPage(lBlock, lCounter - 1))
                    If lCounter < 16384 Then
                       sData = sData & Chr$(gRAMPage(lBlock, lCounter))
                        lCounter = lCounter + 1
                    End If

                Else
                    ' // Strings of 4 or fewer bytes are not encoded
                    sData = sData & String$(lByteCount, Chr$(gRAMPage(lBlock, lCounter - 1)))
                End If
            End If
        Else
            ' // A single byte, just write it out
            sData = sData & Chr$(gRAMPage(lBlock, lCounter))
            lCounter = lCounter + 1
        End If
    Loop Until lCounter >= 16383
    
    If lCounter = 16383 Then
        ' // Write out the final byte
        sData = sData & Chr$(gRAMPage(lBlock, 16383))
    End If
    
    CompressMemoryBlock = Len(sData)
End Function

Public Sub CreateScreenBuffer()
    #If USE_WINAPI Then
        bmiBuffer.bmiHeader.biSize = Len(bmiBuffer.bmiHeader)
        bmiBuffer.bmiHeader.biWidth = 256
        bmiBuffer.bmiHeader.biHeight = 192
        bmiBuffer.bmiHeader.biPlanes = 1
        bmiBuffer.bmiHeader.biBitCount = 8
        bmiBuffer.bmiHeader.biCompression = BI_RGB
        bmiBuffer.bmiHeader.biSizeImage = 0
        bmiBuffer.bmiHeader.biXPelsPerMeter = 200
        bmiBuffer.bmiHeader.biYPelsPerMeter = 200
        bmiBuffer.bmiHeader.biClrUsed = 16
        bmiBuffer.bmiHeader.biClrImportant = 16
    #End If
End Sub

Private Function GetBorderIndex(lRGB As Long) As Long
    Dim lCounter As Long
    
    For lCounter = 0 To 7
        If glNormalColor(lCounter) = lRGB Then
            GetBorderIndex = lCounter
        End If
    Next lCounter
End Function

Function GetFilePart(ByVal sFilename As String) As String
    If InStr(sFilename, ":") > 0 Then sFilename = Mid$(sFilename, InStr(sFilename, ":") + 1)
    Do While InStr(sFilename, "\") > 0
        sFilename = Mid$(sFilename, InStr(sFilename, "\") + 1)
    Loop
    GetFilePart = sFilename
End Function

Public Function InitializeWaveOut() As Boolean
    #If USE_WINAPI Then
        Dim lRet As Long, sErrMsg As String, lCounter As Long, lCounter2 As Long
        
        If val(GetSetting("Grok", "vbSpec", "SoundEnabled", CStr(1))) = 0 Then
            InitializeWaveOut = False
            Exit Function
        End If
        
        With gtWavFormat
            .wFormatTag = WAVE_FORMAT_PCM
            .nChannels = 1
            .nSamplesPerSec = WAVE_FREQUENCY
            .nAvgBytesPerSec = WAVE_FREQUENCY
            .nBlockAlign = 1
            .wBitsPerSample = 8
            .cbSize = 0
        End With
        lRet = waveOutOpen(glphWaveOut, WAVE_MAPPER, gtWavFormat, 0, True, CALLBACK_NULL)
        If lRet <> MMSYSERR_NOERROR Then
            sErrMsg = Space$(255)
            waveOutGetErrorText lRet, sErrMsg, Len(sErrMsg)
            sErrMsg = Left$(sErrMsg, InStr(sErrMsg, Chr$(0)) - 1)
            MsgBox "Error initialising WaveOut device." & vbCrLf & vbCrLf & sErrMsg, vbOKOnly Or vbExclamation, "vbSpec"
            InitializeWaveOut = False
            Exit Function
        End If
        
        For lCounter = 1 To NUM_WAV_BUFFERS
            ghMem(lCounter) = GlobalAlloc(GPTR, WAV_BUFFER_SIZE)
            gpMem(lCounter) = GlobalLock(ghMem(lCounter))
            With gtWavHdr(lCounter)
                .lpData = gpMem(lCounter)
                .dwBufferLength = WAV_BUFFER_SIZE
                .dwUser = 0
                .dwFlags = 0
                .dwLoops = 0
                .lpNext = 0
            End With
            
            lRet = waveOutPrepareHeader(glphWaveOut, gtWavHdr(lCounter), Len(gtWavHdr(lCounter)))
            If lRet <> MMSYSERR_NOERROR Then
                sErrMsg = Space$(255)
                waveOutGetErrorText lRet, sErrMsg, Len(sErrMsg)
                sErrMsg = Left$(sErrMsg, InStr(sErrMsg, Chr$(0)) - 1)
                MsgBox "Error preparing wave header." & vbCrLf & vbCrLf & sErrMsg, vbOKOnly Or vbExclamation, "vbSpec"
                lRet = waveOutClose(glphWaveOut)
                For lCounter2 = 1 To NUM_WAV_BUFFERS
                    GlobalUnlock ghMem(lCounter2)
                    GlobalFree ghMem(lCounter2)
                Next lCounter2
                InitializeWaveOut = False
                Exit Function
            End If
        Next lCounter
        
        InitializeWaveOut = True
    #End If
    
    
End Function

Sub initParity()
    Dim lCounter As Long, j As Byte, p As Boolean
    
    For lCounter = 0 To 255
        p = True
        For j = 0 To 7
            If (lCounter And (2 ^ j)) <> 0 Then p = Not p
        Next j
        Parity(lCounter) = p
    Next lCounter
End Sub

Sub initscreen()
    Dim i As Long, X As Long
    
    ScrnNeedRepaint = True
    For i = 0 To 191
        ScrnLines(i, 32) = True
        For X = 0 To 31
            ScrnLines(i, X) = True
        Next X
    Next i
End Sub

Public Sub InitScreenMemTable()
    Dim X As Long, Y As Long
    
    For Y = 0 To 191
        For X = 0 To 31
            glScreenMem(Y, X) = ((((Y \ 8) * 32) + (Y Mod 8) * 256) + ((Y \ 64) * 2048) - (Y \ 64) * 256) + X
        Next X
    Next Y
End Sub


Sub LoadROM(Optional sROMFile As String = "spectrum.rom", Optional lROMPage As Long = 8)
    Dim hFile As Long, sROM As String, lCounter As Long
    
    On Error Resume Next
    
    If Dir$(sROMFile) = "" Then
        MsgBox "The ROM image file '" & sROMFile & "' could not be found.", vbExclamation Or vbOKOnly
        Exit Sub
    End If
    
    err.Clear
    hFile = FreeFile
    Open sROMFile For Binary As hFile
    If err.Number <> 0 Then
        MsgBox "Unable to open ROM image file: " & sROMFile, vbExclamation Or vbOKOnly
        Close hFile
        Exit Sub
    End If
    
    ' // Read the ROM image into sROM
    err.Clear
    sROM = Input(16384, #hFile)
    Close hFile
        
    If err.Number <> 0 Then
        MsgBox "An error ocurred whilst reading the ROM image file: " & sROMFile, vbExclamation Or vbOKOnly
        Exit Sub
    End If
    
    ' // Copy the ROM into the appropriate memory page
    For lCounter = 1 To 16384
        gRAMPage(lROMPage, lCounter - 1) = Asc(Mid$(sROM, lCounter, 1))
    Next lCounter
        
    
    ' // Copy the ROM in to the low 16KB of addressable memory, and also place a
    ' // shadow copy at 65536 - this means we don't have to check for regPC > 65535
    ' // which improves performance
'    If bPagedIn Then
'        For lCounter = 1 To 16384
'            mem(lCounter - 1) = Asc(Mid$(sROM, lCounter, 1))
'            mem(lCounter + 65535) = mem(lCounter - 1)
'        Next lCounter
'    End If
End Sub

Private Sub LoadSNA128Snap(hFile As Long)
    Dim sData As String, sTemp As String, lOut7FFD As Long, lBank As Long, lCounter As Long
    
    ' // Read first three banks
    sData = Input(49152, #hFile)
    
    ' // PC
    sTemp = Input(2, #hFile)
    regPC = Asc(Right$(sTemp, 1)) * 256& + Asc(Left$(sTemp, 1))
    
    ' // Last out to 0x7FFD
    lOut7FFD = Asc(Input(1, #hFile))
    
    ' // Is TR-DOS paged? (ignored by vbSpec)
    sTemp = Input(1, #hFile)
    
    ' Setup first three banks
    For lCounter = 0 To 16383
        gRAMPage(5, lCounter) = Asc(Mid$(sData, lCounter + 1, 1))
        gRAMPage(2, lCounter) = Asc(Mid$(sData, lCounter + 16385, 1))
        gRAMPage(lOut7FFD And 7, lCounter) = Asc(Mid$(sData, lCounter + 32769, 1))
    Next lCounter
    
    lBank = 0
    Do While lBank < 8
        sData = Input(16384, #hFile)
        If (lBank = 5) Or (lBank = 2) Or (lBank = (lOut7FFD And 7)) Then
            lBank = lBank + 1
        End If
        If lBank < 8 Then
            For lCounter = 0 To 16383
                gRAMPage(lBank, lCounter) = Asc(Mid$(sData, lCounter + 1, 1))
            Next lCounter
            lBank = lBank + 1
        End If
    Loop
    
    outb &H7FFD&, lOut7FFD
End Sub

Public Sub LoadSNASnap(sFilename As String)
    Dim hFile As Long, sData As String, iCounter As Long
    
    hFile = FreeFile
    Open sFilename For Binary As hFile
    
    sData = Input(1, #hFile)
    intI = Asc(sData)
    sData = Input(2, #hFile)
    regHL_ = Asc(Right$(sData, 1)) * 256& + Asc(Left$(sData, 1))
    sData = Input(2, #hFile)
    regDE_ = Asc(Right$(sData, 1)) * 256& + Asc(Left$(sData, 1))
    sData = Input(2, #hFile)
    regBC_ = Asc(Right$(sData, 1)) * 256& + Asc(Left$(sData, 1))
    sData = Input(2, #hFile)
    regAF_ = Asc(Right$(sData, 1)) * 256& + Asc(Left$(sData, 1))
    
    sData = Input(2, #hFile)
    regHL = Asc(Right$(sData, 1)) * 256& + Asc(Left$(sData, 1))
    sData = Input(2, #hFile)
    regDE = Asc(Right$(sData, 1)) * 256& + Asc(Left$(sData, 1))
    sData = Input(2, #hFile)
    setBC Asc(Right$(sData, 1)) * 256& + Asc(Left$(sData, 1))
    sData = Input(2, #hFile)
    regIY = Asc(Right$(sData, 1)) * 256& + Asc(Left$(sData, 1))
    sData = Input(2, #hFile)
    regIX = Asc(Right$(sData, 1)) * 256& + Asc(Left$(sData, 1))
    
    sData = Input(1, #hFile)
    If Asc(sData) And 4 Then
        intIFF1 = True
        intIFF2 = True
    Else
        intIFF1 = False
        intIFF2 = False
    End If
    
    sData = Input(1, #hFile)
    intR = Asc(sData)
    intRTemp = intR
    
    sData = Input(2, #hFile)
    setAF Asc(Right$(sData, 1)) * 256& + Asc(Left$(sData, 1))
    
    sData = Input(2, #hFile)
    regSP = Asc(Right$(sData, 1)) * 256& + Asc(Left$(sData, 1))

    sData = Input(1, #hFile)
    intIM = Asc(sData)
    
    sData = Input(1, #hFile)
    ' // Border color
    glNewBorder = (Asc(sData) And &H7&)
    
    If LOF(hFile) > 49180 Then
        LoadSNA128Snap hFile
        Close hFile
        
        initscreen
    
        ' // Set the initial border color
        frmMainWnd.BackColor = glNormalColor(glNewBorder)
    
        If glEmulatedModel = 0 Then SetEmulatedModel 1
    
        frmMainWnd.Caption = App.ProductName & " - " & GetFilePart(sFilename)
    
        screenPaint
        resetKeyboard
    Else
        ' // Load a 48K Snapshot file
        err.Clear
        sData = Input(49153, #hFile)
        Close hFile
       
        For iCounter = 0 To 16383
            gRAMPage(5, iCounter) = Asc(Mid$(sData, iCounter + 1, 1))
            gRAMPage(1, iCounter) = Asc(Mid$(sData, iCounter + 16385, 1))
            gRAMPage(2, iCounter) = Asc(Mid$(sData, iCounter + 32769, 1))
        Next iCounter
    
        initscreen
    
        ' // Set the initial border color
        frmMainWnd.BackColor = glNormalColor(glNewBorder)
    
        SetEmulatedModel 0
    
        frmMainWnd.Caption = App.ProductName & " - " & GetFilePart(sFilename)
    
        screenPaint
        resetKeyboard
        poppc
    End If
End Sub


Public Sub LoadZ80Snap(sFilename As String)
    Dim hFile As Long, sData As String, iCounter As Long
    Dim bCompressed As Boolean
    
    hFile = FreeFile
    Open sFilename For Binary As hFile
    
    glPageAt(0) = 8
    glPageAt(1) = 5
    glPageAt(2) = 2
    glPageAt(3) = 0
    glPageAt(4) = 8
    
    #If USE_WINAPI Then
        If gbSoundEnabled Then AY8912_reset
    #End If
    
    ' byte 0 - A register
    sData = Input(1, #hFile)
    regA = Asc(sData)
    ' byte 1 - F register
    sData = Input(1, #hFile)
    setF Asc(sData)
    ' bytes 2 + 3 - BC register pair (C first, then B)
    sData = Input(2, #hFile)
    setBC Asc(Right$(sData, 1)) * 256& + Asc(Left$(sData, 1))
    ' bytes 4 + 5 - HL register pair
    sData = Input(2, #hFile)
    regHL = Asc(Right$(sData, 1)) * 256& + Asc(Left$(sData, 1))
    ' bytes 6 + 7 - PC (this is zero for v2.x or v3.0 Z80 files)
    sData = Input(2, #hFile)
    regPC = Asc(Right$(sData, 1)) * 256& + Asc(Left$(sData, 1))
    ' bytes 8 + 9 - SP
    sData = Input(2, #hFile)
    regSP = Asc(Right$(sData, 1)) * 256& + Asc(Left$(sData, 1))
    ' byte 10 - Interrupt register
    sData = Input(1, #hFile)
    intI = Asc(sData)
    ' byte 11 - Refresh register
    sData = Input(1, #hFile)
    intR = (Asc(sData) And 127)
    
    ' byte 12 - bitfield
    sData = Input(1, #hFile)
    ' if byte 12 = 255 then it must be treated as if it = 1, for compatibility with other emulators
    If Asc(sData) = 255 Then sData = Chr$(1)
    ' bit 0 - bit 7 of R
    If (Asc(sData) And 1) = 1 Then intR = intR Or 128
    intRTemp = intR
    ' bits 1,2 and 3 - border color
    glNewBorder = (Asc(sData) And 14) \ 2
    ' bit 4 - 1 if SamROM switched in (we don't care about this!)
    ' bit 5 - if 1 and PC<>0 then the snapshot is compressed using the
    '         rudimentary Z80 run-length encoding scheme
    If (Asc(sData) And &H20&) Then bCompressed = True
    ' bits 6 + 7 - no meaning

    ' bytes 13 + 14 - DE register pair
    sData = Input(2, #hFile)
    regDE = Asc(Right$(sData, 1)) * 256& + Asc(Left$(sData, 1))
    ' bytes 15 + 16 - BC' register pair
    sData = Input(2, #hFile)
    regBC_ = Asc(Right$(sData, 1)) * 256& + Asc(Left$(sData, 1))
    ' bytes 17 + 18 - DE' register pair
    sData = Input(2, #hFile)
    regDE_ = Asc(Right$(sData, 1)) * 256& + Asc(Left$(sData, 1))
    ' bytes 19 + 20 - HL' register pair
    sData = Input(2, #hFile)
    regHL_ = Asc(Right$(sData, 1)) * 256& + Asc(Left$(sData, 1))
    ' bytes 21 + 22 - AF' register pair (A first then F - not Z80 byte order!!)
    sData = Input(2, #hFile)
    regAF_ = Asc(Left$(sData, 1)) * 256& + Asc(Right$(sData, 1))
    ' byte 23 + 24 - IY register pair
    sData = Input(2, #hFile)
    regIY = Asc(Right$(sData, 1)) * 256& + Asc(Left$(sData, 1))
    ' byte 25 + 26 - IX register pair
    sData = Input(2, #hFile)
    regIX = Asc(Right$(sData, 1)) * 256& + Asc(Left$(sData, 1))
    ' byte 27 - Interrupt flipflop (0=DI, else EI)
    sData = Input(1, #hFile)
    If Asc(sData) = 0 Then
        intIFF1 = False
        intIFF2 = False
    Else
        intIFF1 = True
        intIFF2 = True
    End If
    ' byte 28 - IFF2 (ignored)
    sData = Input(1, #hFile)
    ' byte 29 - Interrupt mode (bits 2 - 7 contain info about joystick modes etc, which we ignore)
    sData = Input(1, #hFile)
    intIM = Asc(sData) And 3
    
    If regPC = 0 Then
        ' This is a V2 or V3 Z80 file
        ReadZ80V2orV3Snap hFile
        Close hFile
    Else
        SetEmulatedModel 0 ' // V1 .Z80 snapshots are all 48K
        ' // PC<>0, so lets check to see if this is a compressed V1 Z80 file
        If bCompressed Then
            ' Uncompress the RAM data
            ReadZ80V1Snap hFile
            Close hFile
        Else
            ' // Uncompressed Z80 file
            sData = Input(49153, #hFile)
            Close hFile
            
            ' // Copy the RAM data to addressable memory space
            For iCounter = 16384 To 65535
                gRAMPage(glPageAt(iCounter \ 16384), iCounter And 16383) = Asc(Mid$(sData, iCounter - 16383, 1))
            Next iCounter
        End If
    End If
    
    initscreen
    
    ' // Set the initial border color
    Select Case glNewBorder
    Case 0
        frmMainWnd.BackColor = 0&
    Case 1
        frmMainWnd.BackColor = RGB(0, 0, 192)
    Case 2
        frmMainWnd.BackColor = RGB(192, 0, 0)
    Case 3
        frmMainWnd.BackColor = RGB(192, 0, 192)
    Case 4
        frmMainWnd.BackColor = RGB(0, 192, 0)
    Case 5
        frmMainWnd.BackColor = RGB(0, 192, 192)
    Case 6
        frmMainWnd.BackColor = RGB(192, 192, 0)
    Case 7
        frmMainWnd.BackColor = RGB(192, 192, 192)
    End Select

    frmMainWnd.Caption = App.ProductName & " - " & GetFilePart(sFilename)
    
    initscreen
    screenPaint
    resetKeyboard
        
    gpicDisplay.REFRESH
    DoEvents
End Sub


Sub Main()
    InitColorArrays
    InitScreenMemTable
      
    frmMainWnd.Show
    
    Set gpicDisplay = frmMainWnd.picDisplay
    
    #If USE_WINAPI Then
        CreateScreenBuffer
        glDisplayWidth = val(GetSetting("Grok", "vbSpec", "DisplayWidth", "256"))
        glDisplayHeight = val(GetSetting("Grok", "vbSpec", "DisplayHeight", "192"))
        SetDisplaySize glDisplayWidth, glDisplayHeight
        AY8912_init 1773000, WAVE_FREQUENCY, 8
    #Else
        ' // Hide the "Display..." Options item, as it can't be changed unless we're
        ' // using the fast WINAPI method of screen painting
        frmMainWnd.mnuOptions(2).Visible = False
    #End If
    
    glInterruptDelay = val(GetSetting("Grok", "vbSpec", "InterruptDelay", "20"))
    
    ' // Initialize everything
    initParity
    Z80Reset
    
    SetEmulatedModel val(GetSetting("Grok", "vbSpec", "EmulatedModel", "0"))
    
    initscreen
    resetKeyboard
    
    timeBeginPeriod 1
    
    glInterruptTimer = timeGetTime() + 20
    #If USE_WINAPI Then
        gbSoundEnabled = InitializeWaveOut()
        If gbSoundEnabled Then
            For glBufNum = 1 To 12
                waveOutWrite glphWaveOut, gtWavHdr(glBufNum), Len(gtWavHdr(glBufNum))
            Next glBufNum
            glBufNum = 12
        End If
    #End If
    
    ' // Begin executing the ROM code at 0x0000
    execute
End Sub

Public Sub InitColorArrays()
#If USE_WINAPI Then
    bmiBuffer.bmiColors(0).rgbRed = 0
    bmiBuffer.bmiColors(0).rgbGreen = 0
    bmiBuffer.bmiColors(0).rgbBlue = 0
    
    bmiBuffer.bmiColors(1).rgbRed = 0
    bmiBuffer.bmiColors(1).rgbGreen = 0
    bmiBuffer.bmiColors(1).rgbBlue = 192
    
    bmiBuffer.bmiColors(2).rgbRed = 192
    bmiBuffer.bmiColors(2).rgbGreen = 0
    bmiBuffer.bmiColors(2).rgbBlue = 0
    
    bmiBuffer.bmiColors(3).rgbRed = 192
    bmiBuffer.bmiColors(3).rgbGreen = 0
    bmiBuffer.bmiColors(3).rgbBlue = 192
    
    bmiBuffer.bmiColors(4).rgbRed = 0
    bmiBuffer.bmiColors(4).rgbGreen = 192
    bmiBuffer.bmiColors(4).rgbBlue = 0
    
    bmiBuffer.bmiColors(5).rgbRed = 0
    bmiBuffer.bmiColors(5).rgbGreen = 192
    bmiBuffer.bmiColors(5).rgbBlue = 192
    
    bmiBuffer.bmiColors(6).rgbRed = 192
    bmiBuffer.bmiColors(6).rgbGreen = 192
    bmiBuffer.bmiColors(6).rgbBlue = 0
    
    bmiBuffer.bmiColors(7).rgbRed = 192
    bmiBuffer.bmiColors(7).rgbGreen = 192
    bmiBuffer.bmiColors(7).rgbBlue = 192
    
    bmiBuffer.bmiColors(8).rgbRed = 0
    bmiBuffer.bmiColors(8).rgbGreen = 0
    bmiBuffer.bmiColors(8).rgbBlue = 0
    
    bmiBuffer.bmiColors(9).rgbRed = 0
    bmiBuffer.bmiColors(9).rgbGreen = 0
    bmiBuffer.bmiColors(9).rgbBlue = 255
    
    bmiBuffer.bmiColors(10).rgbRed = 255
    bmiBuffer.bmiColors(10).rgbGreen = 0
    bmiBuffer.bmiColors(10).rgbBlue = 0
    
    bmiBuffer.bmiColors(11).rgbRed = 255
    bmiBuffer.bmiColors(11).rgbGreen = 0
    bmiBuffer.bmiColors(11).rgbBlue = 255
    
    bmiBuffer.bmiColors(12).rgbRed = 0
    bmiBuffer.bmiColors(12).rgbGreen = 255
    bmiBuffer.bmiColors(12).rgbBlue = 0
    
    bmiBuffer.bmiColors(13).rgbRed = 0
    bmiBuffer.bmiColors(13).rgbGreen = 255
    bmiBuffer.bmiColors(13).rgbBlue = 255
    
    bmiBuffer.bmiColors(14).rgbRed = 255
    bmiBuffer.bmiColors(14).rgbGreen = 255
    bmiBuffer.bmiColors(14).rgbBlue = 0
    
    bmiBuffer.bmiColors(15).rgbRed = 255
    bmiBuffer.bmiColors(15).rgbGreen = 255
    bmiBuffer.bmiColors(15).rgbBlue = 255
#End If
    glBrightColor(0) = 0
    glBrightColor(1) = RGB(0, 0, 255)
    glBrightColor(2) = RGB(255, 0, 0)
    glBrightColor(3) = RGB(255, 0, 255)
    glBrightColor(4) = RGB(0, 255, 0)
    glBrightColor(5) = RGB(0, 255, 255)
    glBrightColor(6) = RGB(255, 255, 0)
    glBrightColor(7) = RGB(255, 255, 255)
    glNormalColor(0) = 0
    glNormalColor(1) = RGB(0, 0, 192)
    glNormalColor(2) = RGB(192, 0, 0)
    glNormalColor(3) = RGB(192, 0, 192)
    glNormalColor(4) = RGB(0, 192, 0)
    glNormalColor(5) = RGB(0, 192, 192)
    glNormalColor(6) = RGB(192, 192, 0)
    glNormalColor(7) = RGB(192, 192, 192)
End Sub



Private Sub ReadZ80V1Snap(hFile As Long)
    Dim lDataLen As Long, sData As String, lBlockLen As Long
    Dim lCounter As Long, lMemPos As Long, lBlockCounter As Long
 
    lDataLen = LOF(hFile) - Seek(hFile) + 1
    ' // read the compressed data into sData
    sData = Input(lDataLen, #hFile)
        
    ' // Uncompress the block to memory
    lCounter = 1
    lMemPos = 16384
    Do
        If Asc(Mid$(sData, lCounter, 1)) = &HED& Then
            If Asc(Mid$(sData, lCounter + 1, 1)) = &HED& Then
                ' // This is an encoded block
                lCounter = lCounter + 2
                lBlockLen = Asc(Mid$(sData, lCounter, 1))
                lCounter = lCounter + 1
                For lBlockCounter = 1 To lBlockLen
                    gRAMPage(glPageAt(lMemPos \ 16384), lMemPos And 16383) = Asc(Mid$(sData, lCounter, 1))
                    lMemPos = lMemPos + 1
                Next lBlockCounter
            Else
                ' // Just a single ED, write it out
                gRAMPage(glPageAt(lMemPos \ 16384), lMemPos And 16383) = &HED&
                lMemPos = lMemPos + 1
            End If
        Else
            gRAMPage(glPageAt(lMemPos \ 16384), lMemPos And 16383) = Asc(Mid$(sData, lCounter, 1))
            lMemPos = lMemPos + 1
        End If
        lCounter = lCounter + 1
    Loop Until lCounter > Len(sData) - 4
    
    If Mid$(sData, lCounter, 4) <> Chr$(0) & Chr$(&HED) & Chr$(&HED) & Chr$(0) Then
        MsgBox "Error in compressed Z80 file. Block end marker 0x00EDED00 is not present."
    End If

End Sub

Private Sub ReadZ80V2orV3Snap(hFile As Long)
Dim lHeaderLen As Long, sData As String
    Dim lCounter As Long, bHardwareSupported As Boolean
    Dim lMemPage As Long, lBlockCounter As Long, b128K As Boolean, lMemPos As Long
    Dim lOutFFFD As Long
    
    sData = Input(2, #hFile)
    lHeaderLen = Asc(Right$(sData, 1)) * 256& + Asc(Left$(sData, 1))
    
    ' // offset 32 - 2 bytes - PC
    If lCounter < lHeaderLen Then
        sData = Input(2, #hFile)
        regPC = Asc(Right$(sData, 1)) * 256& + Asc(Left$(sData, 1))
        lCounter = lCounter + 2
    End If
    
    ' // offset 34 - 1 byte - hardware mode
    If lCounter < lHeaderLen Then
        sData = Input(1, #hFile)
        Select Case Asc(sData)
        Case 0 ' // 48K spectrum
            bHardwareSupported = True
            SetEmulatedModel 0
        Case 1 ' // 48K spectrum + Interface 1
            bHardwareSupported = True
            SetEmulatedModel 0
        Case 2 ' // SamROM
            sData = "SamROM"
            bHardwareSupported = False
        Case 3
            If lHeaderLen = 23 Then
                sData = "128K Spectrum"
                bHardwareSupported = True
                b128K = True
                If glEmulatedModel = 0 Then SetEmulatedModel 1
            Else
                sData = "48K Spectrum + M.G.T."
                bHardwareSupported = True
                SetEmulatedModel 0
            End If
        Case 4
            If lHeaderLen = 23 Then
                sData = "128K Spectrum + Interface 1"
                bHardwareSupported = True
                b128K = True
                If glEmulatedModel = 0 Then SetEmulatedModel 1
            Else
                sData = "128K Spectrum"
                bHardwareSupported = True
                b128K = True
                If glEmulatedModel = 0 Then SetEmulatedModel 1
            End If
        Case 5
            bHardwareSupported = True
            sData = "128K Spectrum + Interface 1"
            b128K = True
            If glEmulatedModel = 0 Then SetEmulatedModel 1
        Case 6
            bHardwareSupported = False
            sData = "128K Spectrum + M.G.T."
            b128K = True
            If glEmulatedModel = 0 Then SetEmulatedModel 1
        Case 7
            bHardwareSupported = False
            sData = "ZX Spectrum +3"
            b128K = True
            If glEmulatedModel = 0 Then SetEmulatedModel 2
        Case Else
            bHardwareSupported = False
            sData = "Unknown hardware platform"
        End Select
        
        If bHardwareSupported = False Then
            MsgBox "vbSpec does not support the required hardware platform (" & sData & ") for this snapshot. This snapshot may not function correctly in vbSpec."
        End If
        
        lCounter = lCounter + 1
    End If
    
    ' // offset 35 - 1 byte - last out to 0x7FFD - not required for 48K spectrum
    If lCounter < lHeaderLen Then
        If b128K Then
            outb &H7FFD&, CLng(Asc(Input(1, #hFile)))
        Else
            sData = Input(1, #hFile)
            glPageAt(0) = 8
            glPageAt(1) = 5
            glPageAt(2) = 1
            glPageAt(3) = 2
            glPageAt(4) = 8
        End If
        lCounter = lCounter + 1
    End If
    
    ' // offset 36 - 1 byte - 0xFF if Interface 1 ROM is paged in
    If lCounter < lHeaderLen Then
        sData = Input(1, #hFile)
        lCounter = lCounter + 1
        
        If Asc(sData) = 255 Then
            MsgBox "This snapshot was saved with the Interface 1 ROM paged in. It may not run correctly in vbSpec."
        End If
    End If
    
    ' // offset 37 - 1 byte (bit 0: 1=intR emulation on, bit 1: 1=LDIR emulation on
    If lCounter < lHeaderLen Then
        sData = Input(1, #hFile)
        lCounter = lCounter + 1
    End If
    
    ' // offset 38 - Last out to 0xFFFD (+2/+3 sound chip register number)
    If lCounter < lHeaderLen Then
        lOutFFFD = Asc(Input(1, #hFile))
        lCounter = lCounter + 1
    End If
    
    ' // offset 39 - 16 bytes - contents of the sound chip registers
    If lCounter < lHeaderLen Then
        If b128K Then
            AYWriteReg 0, Asc(Input(1, #hFile))
            AYWriteReg 1, Asc(Input(1, #hFile))
            AYWriteReg 2, Asc(Input(1, #hFile))
            AYWriteReg 3, Asc(Input(1, #hFile))
            AYWriteReg 4, Asc(Input(1, #hFile))
            AYWriteReg 5, Asc(Input(1, #hFile))
            AYWriteReg 6, Asc(Input(1, #hFile))
            AYWriteReg 7, Asc(Input(1, #hFile))
            AYWriteReg 8, Asc(Input(1, #hFile))
            AYWriteReg 9, Asc(Input(1, #hFile))
            AYWriteReg 10, Asc(Input(1, #hFile))
            AYWriteReg 11, Asc(Input(1, #hFile))
            AYWriteReg 12, Asc(Input(1, #hFile))
            AYWriteReg 13, Asc(Input(1, #hFile))
            AYWriteReg 14, Asc(Input(1, #hFile))
            AYWriteReg 15, Asc(Input(1, #hFile))
        Else
            sData = Input(16, #hFile)
        End If
        lCounter = lCounter + 16
    End If
    
    If b128K Then
        outb &HFFFD&, lOutFFFD
    End If
    
    ' // read the remaining bytes of the header (we don't care what information they hold)
    If lCounter < lHeaderLen Then
        sData = Input(lHeaderLen - lCounter, #hFile)
    End If
    
    Do
        ' // read a block
        sData = Input(2, #hFile)
        If EOF(hFile) Then Exit Do
        lHeaderLen = Asc(Right$(sData, 1)) * 256& + Asc(Left$(sData, 1))
        sData = Input(1, #hFile)
        Select Case Asc(sData)
        Case 0 ' // Spectrum ROM
            If b128K Then lMemPage = 9 Else lMemPage = 8
        Case 1 ' // Interface 1 ROM, or similar (we discard these blocks)
            lMemPage = -1
        Case 2 ' // 128K ROM (reset)
            If b128K Then lMemPage = 8 Else lMemPage = -1
        Case 3 ' // Page 0 (not used by 48K Spectrum)
            If b128K Then lMemPage = 0 Else lMemPage = -1
        Case 4 ' // Page 1 RAM at 0x8000
            lMemPage = 1
        Case 5 ' // Page 2 RAM at 0xC000
            lMemPage = 2
        Case 6 ' // Page 3 (not used by 48K Spectrum)
            If b128K Then lMemPage = 3 Else lMemPage = -1
        Case 7 ' // Page 4 (not used by 48K Spectrum)
            If b128K Then lMemPage = 4 Else lMemPage = -1
        Case 8 ' // Page 5 RAM at 0x4000
            lMemPage = 5
        Case 9 ' // Page 6 (not used by 48K Spectrum)
            If b128K Then lMemPage = 6 Else lMemPage = -1
        Case 10 ' // Page 7 (not used by 48K Spectrum)
            If b128K Then lMemPage = 7 Else lMemPage = -1
        Case 11 ' // Multiface ROM
        End Select
        sData = Input(lHeaderLen, #hFile)
        
        If lMemPage <> -1 Then
            If lHeaderLen = &HFFFF& Then
                ' Not a compressed block, just copy it straight into RAM
                For lCounter = 0 To 16383
                    gRAMPage(lMemPage, lCounter) = Asc(Mid$(sData, lCounter + 1, 1))
                Next lCounter
            Else
                ' // Uncompress the block to memory
                lCounter = 1
                lMemPos = 0
                Do
                    If Asc(Mid$(sData, lCounter, 1)) = &HED& Then
                        If Asc(Mid$(sData, lCounter + 1, 1)) = &HED& Then
                            ' // This is an encoded block
                            lCounter = lCounter + 2
                            lHeaderLen = Asc(Mid$(sData, lCounter, 1))
                            lCounter = lCounter + 1
                            For lBlockCounter = 0 To lHeaderLen - 1
                                gRAMPage(lMemPage, lMemPos + lBlockCounter) = Asc(Mid$(sData, lCounter, 1))
                            Next lBlockCounter
                            lMemPos = lMemPos + lBlockCounter
                        Else
                            ' // Just a single ED, write it out
                            gRAMPage(lMemPage, lMemPos) = &HED&
                            lMemPos = lMemPos + 1
                        End If
                    Else
                        gRAMPage(lMemPage, lMemPos) = Asc(Mid$(sData, lCounter, 1))
                        lMemPos = lMemPos + 1
                    End If
                    lCounter = lCounter + 1
                Loop Until lCounter > Len(sData)
            End If
        End If
    Loop Until EOF(hFile)
End Sub

Sub resetKeyboard()
    keyB_SPC = &HFF&
    keyH_ENT = &HFF&
    keyY_P = &HFF&
    key6_0 = &HFF&
    key1_5 = &HFF&
    keyQ_T = &HFF&
    keyA_G = &HFF&
    keyCAPS_V = &HFF&
End Sub


Public Sub SaveROM(sFilename As String)
    Dim hFile As Long, lCounter As Long
    
    On Error GoTo SaveROM_Err
    
    hFile = FreeFile
    Open sFilename For Output As hFile
    For lCounter = 0 To 16383
        Print #hFile, Chr$(gRAMPage(glPageAt(0), lCounter));
    Next lCounter
    
SaveROM_Err:
    Close hFile
End Sub

Private Sub SaveSNA128Snap(hFile As Long)
    Dim lBank As Long, lCounter As Long
    
    Print #hFile, Chr$(intI);
    Print #hFile, Chr$(regHL_ And &HFF&); Chr$(regHL_ \ 256&);
    Print #hFile, Chr$(regDE_ And &HFF&); Chr$(regDE_ \ 256&);
    Print #hFile, Chr$(regBC_ And &HFF&); Chr$(regBC_ \ 256&);
    Print #hFile, Chr$(regAF_ And &HFF&); Chr$(regAF_ \ 256&);
    
    Print #hFile, Chr$(regHL And &HFF&); Chr$(regHL \ 256&);
    Print #hFile, Chr$(regDE And &HFF&); Chr$(regDE \ 256&);
    Print #hFile, Chr$(regC); Chr$(regB);
    Print #hFile, Chr$(regIY And &HFF&); Chr$(regIY \ 256&);
    Print #hFile, Chr$(regIX And &HFF&); Chr$(regIX \ 256&);
    
    ' Interrupt flipflops
    If intIFF1 = True Then
        Print #hFile, Chr$(4);
    Else
        Print #hFile, Chr$(0);
    End If

    ' R
    intRTemp = intRTemp And 127
    Print #hFile, Chr$((intR And &H80&) Or intRTemp);

    ' // AF
    Print #hFile, Chr$(getAF And &HFF&); Chr$(getAF \ 256&);
    
    ' // SP
    Print #hFile, Chr$(regSP And &HFF&); Chr$(regSP \ 256&);
    
    ' // Interrupt Mode
    Print #hFile, Chr$(intIM);
    
    Print #hFile, Chr$(GetBorderIndex(frmMainWnd.BackColor));

    ' // Save the three currently-paged RAM banks
    For lCounter = 0 To 16383
        Print #hFile, Chr$(gRAMPage(glPageAt(1), lCounter));
    Next lCounter
    For lCounter = 0 To 16383
        Print #hFile, Chr$(gRAMPage(glPageAt(2), lCounter));
    Next lCounter
    For lCounter = 0 To 16383
        Print #hFile, Chr$(gRAMPage(glPageAt(3), lCounter));
    Next lCounter

    ' // PC
    Print #hFile, Chr$(regPC And &HFF&); Chr$(regPC \ 256&);
    
    ' // Last out to 0x7FFD
    Print #hFile, Chr$(glLastOut7FFD);
    
    ' // Is TR-DOS paged? (0=not paged, 1=paged)
    Print #hFile, Chr$(0);
    
    ' // Save the remaining RAM banks
    lBank = 0
    Do While lBank < 8
        If lBank <> glPageAt(1) And lBank <> glPageAt(2) And lBank <> glPageAt(3) Then
            For lCounter = 0 To 16383
                Print #hFile, Chr$(gRAMPage(lBank, lCounter));
            Next lCounter
        End If
        lBank = lBank + 1
    Loop
End Sub

Public Sub SaveSNASnap(sFilename As String)
    Dim hFile As Long, sData As String, lCounter As Long
    
    hFile = FreeFile
    Open sFilename For Output As hFile
    
    If glEmulatedModel <> 0 Then
        ' // We're running in 128 mode
        SaveSNA128Snap hFile
        Close hFile
        Exit Sub
    End If
    pushpc
    
    Print #hFile, Chr$(intI);
    Print #hFile, Chr$(regHL_ And &HFF&); Chr$(regHL_ \ 256&);
    Print #hFile, Chr$(regDE_ And &HFF&); Chr$(regDE_ \ 256&);
    Print #hFile, Chr$(regBC_ And &HFF&); Chr$(regBC_ \ 256&);
    Print #hFile, Chr$(regAF_ And &HFF&); Chr$(regAF_ \ 256&);
    
    Print #hFile, Chr$(regHL And &HFF&); Chr$(regHL \ 256&);
    Print #hFile, Chr$(regDE And &HFF&); Chr$(regDE \ 256&);
    Print #hFile, Chr$(regC); Chr$(regB);
    Print #hFile, Chr$(regIY And &HFF&); Chr$(regIY \ 256&);
    Print #hFile, Chr$(regIX And &HFF&); Chr$(regIX \ 256&);
    
    ' Interrupt flipflops
    If intIFF1 = True Then
        Print #hFile, Chr$(4);
    Else
        Print #hFile, Chr$(0);
    End If

    ' R
    intRTemp = intRTemp And 127
    Print #hFile, Chr$((intR And &H80&) Or intRTemp);

    ' // AF
    Print #hFile, Chr$(getAF And &HFF&); Chr$(getAF \ 256&);
    
    ' // SP
    Print #hFile, Chr$(regSP And &HFF&); Chr$(regSP \ 256&);
    
    ' // Interrupt Mode
    Print #hFile, Chr$(intIM);
    
    Print #hFile, Chr$(GetBorderIndex(frmMainWnd.BackColor));

    For lCounter = 0 To 16383
        Print #hFile, Chr$(gRAMPage(glPageAt(1), lCounter));
    Next lCounter
    For lCounter = 0 To 16383
        Print #hFile, Chr$(gRAMPage(glPageAt(2), lCounter));
    Next lCounter
    For lCounter = 0 To 16383
        Print #hFile, Chr$(gRAMPage(glPageAt(3), lCounter));
    Next lCounter
    
    Close hFile
    poppc
End Sub

Public Sub SaveZ80Snap(sFilename As String)
    Dim hFile As Long, sData As String, lCounter As Long, lBufSize As Long
    
    hFile = FreeFile
    Open sFilename For Output As hFile
    
    ' A,F
    Print #hFile, Chr$(regA); Chr$(getF);
    ' BC
    Print #hFile, Chr$(regC); Chr$(regB);
    ' HL
    Print #hFile, Chr$(regHL And &HFF&); Chr$(regHL \ 256&);
    ' Set PC to zero to indicate a v3.0 Z80 file
    Print #hFile, Chr$(0); Chr$(0);
    ' SP
    Print #hFile, Chr$(regSP And &HFF&); Chr$(regSP \ 256&);
    ' I
    Print #hFile, Chr$(intI);
    ' R (7 bits)
    intRTemp = intRTemp And &H7F&
    Print #hFile, Chr$(intRTemp);
    ' bitfield
    Print #hFile, Chr$((IIf(intR And &H80& = &H80, 1, 0)) Or (GetBorderIndex(frmMainWnd.BackColor) * 2));
    ' DE
    Print #hFile, Chr$(regDE And &HFF&); Chr$(regDE \ 256&);
    ' BC'
    Print #hFile, Chr$(regBC_ And &HFF&); Chr$(regBC_ \ 256&);
    ' DE'
    Print #hFile, Chr$(regDE_ And &HFF&); Chr$(regDE_ \ 256&);
    ' HL'
    Print #hFile, Chr$(regHL_ And &HFF&); Chr$(regHL_ \ 256&);
    ' AF'
    Print #hFile, Chr$(regAF_ \ 256&); Chr$(regAF_ And &HFF&);
    ' IY
    Print #hFile, Chr$(regIY And &HFF&); Chr$(regIY \ 256&);
    ' IX
    Print #hFile, Chr$(regIX And &HFF&); Chr$(regIX \ 256&);
    ' Interrupt flipflops
    If intIFF1 = True Then
        Print #hFile, Chr$(255);
        Print #hFile, Chr$(255);
    Else
        Print #hFile, Chr$(0);
        Print #hFile, Chr$(0);
    End If
    ' // Interrupt Mode
    Print #hFile, Chr$(intIM);
    
    ' // V2.01 info
    Print #hFile, Chr$(23) & Chr$(0);
    ' PC
    Print #hFile, Chr$(regPC And &HFF&); Chr$(regPC \ 256&);
    ' Hardware mode
    If glEmulatedModel = 0 Then
        Print #hFile, Chr$(0); ' // 48K Spectrum
        Print #hFile, Chr$(0);
    Else
        Print #hFile, Chr$(3); ' // 128K Spectrum
        ' Last out to 7FFD
        Print #hFile, Chr$(glLastOut7FFD);
    End If
    ' IF.1 Paged in
    Print #hFile, Chr$(0);
    ' 1=R emulation on,2=LDIR emulation on
    Print #hFile, Chr$(3);
    ' Last out to FFFD
    #If USE_WINAPI Then
        Print #hFile, Chr$(glSoundRegister);
    #Else
        Print #hFile, Chr$(0);
    #End If
    ' AY-3-8912 register contents
    For lCounter = 0 To 15
        Print #hFile, Chr$(AYPSG.Regs(lCounter));
    Next lCounter
    
    If glEmulatedModel = 0 Then
        ' // Block 1
        sData = ""
        lBufSize = CompressMemoryBlock(glPageAt(1), sData)
        ' Buffer length
        Print #hFile, Chr$(lBufSize And &HFF&); Chr$(lBufSize \ 256&);
        ' Block number
        Print #hFile, Chr$(glPageAt(1) + 3);
        Print #hFile, sData;
        ' // Block 2
        sData = ""
        lBufSize = CompressMemoryBlock(glPageAt(2), sData)
        ' Buffer length
        Print #hFile, Chr$(lBufSize And &HFF&); Chr$(lBufSize \ 256&);
        ' Block number
        Print #hFile, Chr$(glPageAt(2) + 3);
        Print #hFile, sData;
        ' // Block 3
        sData = ""
        lBufSize = CompressMemoryBlock(glPageAt(3), sData)
        ' Buffer length
        Print #hFile, Chr$(lBufSize And &HFF&); Chr$(lBufSize \ 256&);
        ' Block number
        Print #hFile, Chr$(glPageAt(3) + 3);
        Print #hFile, sData;
    Else
        For lCounter = 0 To 7
            sData = ""
            lBufSize = CompressMemoryBlock(lCounter, sData)
            ' Buffer length
            Print #hFile, Chr$(lBufSize And &HFF&); Chr$(lBufSize \ 256&);
            ' Block number
            Print #hFile, Chr$(lCounter + 3);
            Print #hFile, sData;
        Next lCounter
    End If
        
    Close hFile
End Sub


Public Sub SetDisplaySize(lWidth As Long, lHeight As Long)
    #If USE_WINAPI Then
        Dim i As Integer, Y As Long, X As Long
        
        Y = (GetSystemMetrics(SM_CYCAPTION) + GetSystemMetrics(SM_CYMENU) + GetSystemMetrics(SM_CYFRAME) * 2 + frmMainWnd.picStatus.Height) * Screen.TwipsPerPixelY
        X = (GetSystemMetrics(SM_CXFRAME) * 2) * Screen.TwipsPerPixelX
        glDisplayWidth = lWidth
        glDisplayHeight = lHeight
        glDisplayVSource = lHeight - 1
        glDisplayVSize = -lHeight
        
        SaveSetting "Grok", "vbSpec", "DisplayWidth", CStr(glDisplayWidth)
        SaveSetting "Grok", "vbSpec", "DisplayHeight", CStr(glDisplayHeight)
        
        Select Case lWidth
        Case 256 ' // Standard
            frmMainWnd.Width = 4200 + X
            glDisplayXMultiplier = 1
        Case 512
            frmMainWnd.Width = 8040 + X
            glDisplayXMultiplier = 2
        Case 768
            frmMainWnd.Width = 11880 + X
            glDisplayXMultiplier = 3
        End Select
            
        Select Case lHeight
        Case 192
            frmMainWnd.Height = 3300 + Y
            glDisplayYMultiplier = 1
        Case 384
            frmMainWnd.Height = 6180 + Y
            glDisplayYMultiplier = 2
        Case 576
            frmMainWnd.Height = 9060 + Y
            glDisplayYMultiplier = 3
        End Select
            
        gpicDisplay.Width = lWidth
        gpicDisplay.Height = lHeight
        
        initscreen
    #End If
End Sub


Public Sub SetEmulatedModel(lModel As Long)
    Dim sModel As String
    
    glEmulatedModel = lModel
    
    Select Case lModel
    Case 0
        ' // A 48K Spectrum has 69888 tstates per interrupt (3.50000 MHz)
        sModel = "ZX Spectrum 48K"
        glTstatesPerInterrupt = 69888
        #If USE_WINAPI Then
            glWaveAddTStates = 158
        #End If
        glMemPagingType = 0
        glUseScreen = 5
        
        glPageAt(0) = 8
        glPageAt(1) = 5
        glPageAt(2) = 1
        glPageAt(3) = 2
        glPageAt(4) = 8
        ' // load the ROM image into memory
        LoadROM App.Path & "\spectrum.rom", 8
    Case 1
        ' // A 128K Spectrum has 70908 tstates per interrupt (3.54690 MHz)
        sModel = "ZX Spectrum 128"
        glTstatesPerInterrupt = 70908
        #If USE_WINAPI Then
            glWaveAddTStates = 160
        #End If
        glMemPagingType = 1
        glUseScreen = 5
        
        glPageAt(0) = 8
        glPageAt(1) = 5
        glPageAt(2) = 2
        glPageAt(3) = 0
        glPageAt(4) = 8
        LoadROM App.Path & "\zx128_1.rom", 9
        LoadROM App.Path & "\zx128_0.rom", 8
    Case 2
        ' // A Spectrum +2 has 70908 tstates per interrupt (3.54690 MHz)
        sModel = "ZX Spectrum +2"
        glTstatesPerInterrupt = 70908
        #If USE_WINAPI Then
            glWaveAddTStates = 160
        #End If
        glMemPagingType = 1
        glUseScreen = 5
        
        glPageAt(0) = 8
        glPageAt(1) = 5
        glPageAt(2) = 2
        glPageAt(3) = 0
        glPageAt(4) = 8
        LoadROM App.Path & "\plus2_1.rom", 9
        LoadROM App.Path & "\plus2_0.rom", 8
    Case 3 ' // +2A
        sModel = "ZX Spectrum +2A"
        glTstatesPerInterrupt = 70908
        #If USE_WINAPI Then
            glWaveAddTStates = 160
        #End If

        glMemPagingType = 2
        glUseScreen = 5
        
        glPageAt(0) = 8
        glPageAt(1) = 5
        glPageAt(2) = 1
        glPageAt(3) = 2
        glPageAt(4) = 8
        
        LoadROM App.Path & "\plus2a_3.rom", 11
        LoadROM App.Path & "\plus2a_2.rom", 10
        LoadROM App.Path & "\plus2a_1.rom", 9
        LoadROM App.Path & "\plus2a_0.rom", 8
    Case 4 ' // +3
        sModel = "ZX Spectrum +3"
        glTstatesPerInterrupt = 70908
        #If USE_WINAPI Then
            glWaveAddTStates = 160
        #End If
        glMemPagingType = 2
        glUseScreen = 5
        
        glPageAt(0) = 8
        glPageAt(1) = 5
        glPageAt(2) = 1
        glPageAt(3) = 2
        glPageAt(4) = 8
        
        LoadROM App.Path & "\plus3_3.rom", 11
        LoadROM App.Path & "\plus3_2.rom", 10
        LoadROM App.Path & "\plus3_1.rom", 9
        LoadROM App.Path & "\plus3_0.rom", 8
    End Select
    
    SetStatus sModel
End Sub


Public Sub SetStatus(sMsg As String)
    frmMainWnd.lblStatus.Caption = sMsg
End Sub


